# This example shows how to use oplog package with FastAPI.
# The get API endpoint is decorated with @Operated, which means that every request 
# will be logged as an operation.
# All subsequent operations will correlated to that operation using the correlation_id.
# To run this example, run the following command from the repository root directory:
# uvicorn examples.server_demo.main:app --reload

import logging

from fastapi import FastAPI

import os
import sys

import uvicorn


# Get the parent directory of the current file (project_demo folder)
current_dir = os.path.dirname(os.path.abspath(__file__))
# Get the path to the repository root directory (one level up)
repository_root = os.path.abspath(os.path.join(current_dir, "..", ".."))

# Add the repository root directory to the Python path
sys.path.append(repository_root)

from oplog import Operation, Operated, OperationHandler # noqa: E402
from oplog.formatters import VerboseOplogLineFormatter # noqa: E402

stream_op_handler = OperationHandler(
    handler=logging.StreamHandler(), # <-- any logging handler
    formatter=VerboseOplogLineFormatter(), # <-- custom formatter or built-in ones
)
logging.basicConfig(level=logging.INFO, handlers=[stream_op_handler])

app = FastAPI()

@app.get("/")
async def read_root():
    with Operation("root"):
        return { "Hello": compute_response() }

@Operated()
def compute_response():
    return "World"

if __name__ == "__main__":
    uvicorn.run(app, host="0.0.0.0", port=8000)





from oplog.operated import Operated
from oplog.tests.logged_test_case import OpLogTestCase


@Operated()
def operated_function():
    pass


class OperatedTestClass:
    @Operated()
    def operated_method_without_name(self):
        pass
    
    @Operated("test_op")
    def operated_method(self):
        pass
    
    @staticmethod
    @Operated("test_op_static")
    def operated_static_method():
        pass

class TestOperated(OpLogTestCase):
    # TODO the correct testing would be to mock Operation and to 
    # make sure the constructor is called with the correct arguments
    
    def test_operated_withoutName_underlyingOperationCreated(self):
        otc = OperatedTestClass()
        otc.operated_method_without_name()
        
        self.assertEqual(len(self.ops), 1)

        op = self.ops[0]
        self.assertEqual(op.name, "OperatedTestClass.operated_method_without_name")

    def test_operated_vanilla_underlyingOperationCreated(self):
        otc = OperatedTestClass()
        otc.operated_method()
        
        self.assertEqual(len(self.ops), 1)

        op = self.ops[0]
        self.assertEqual(op.name, "test_op")
    
    def test_operated_staticMethod_underlyingOperationCreated(self):
        otc = OperatedTestClass()
        otc.operated_static_method()
        
        self.assertEqual(len(self.ops), 1)

        op = self.ops[0]
        self.assertEqual(op.name, "test_op_static")
        
    def test_operated_function_underlyingOperationCreated(self):
        operated_function()
        
        self.assertEqual(len(self.ops), 1)

        op = self.ops[0]
        self.assertEqual(op.name, "operated_function.operated_function")



import logging
from unittest import mock
from oplog import OperationHandler
from oplog.operation import Operation
from oplog.tests.logged_test_case import OpLogTestCase


class TestOperationHandler(OpLogTestCase):
    def test_operationHandler_loggingOperation_operationLogged(self):
        handler = OperationHandler(logging.StreamHandler())
        logging.getLogger().addHandler(handler)
        
        with Operation(name="test_op") as op:
            pass
        
        self.assertEqual(self.get_op("test_op"), op)
        
    def test_operationHandler_loggingOperationAndInfo_infoNotLoggedByHandler(self):
        handler_mock = mock.Mock()
        op_handler = OperationHandler(handler_mock)
        logging.getLogger().addHandler(op_handler)
        
        with Operation(name="test_op") as op:
            pass
            logging.getLogger().info("test_info")
        
        # assert called once    
        self.assertEqual(len(handler_mock.emit.call_args_list), 1)
        
        # assert called once with a log record with op
        call_args = handler_mock.emit.call_args_list[0][0]
        self.assertEqual(call_args[0].oplog, op)

    def test_operationHandler_loggingInfo_infoNotLoggedByHandler(self):
        handler_mock = mock.Mock()
        op_handler = OperationHandler(handler_mock)
        logging.getLogger().addHandler(op_handler)
        
        logging.getLogger().info("test_info")
        
        # assert called once    
        handler_mock.emit.assert_not_called()



import logging
from unittest.mock import Mock
from oplog import OperationLogFilter
from oplog.tests.logged_test_case import OpLogTestCase


class OperationExceptionTest(Exception):
    pass


class TestOperationLogFilter(OpLogTestCase):
    def test_filter_recordIsVanillaLogRecord_false(self):
        mock_record = Mock(spec=logging.LogRecord)
        filter = OperationLogFilter()
        self.assertFalse(filter.filter(record=mock_record))
        
    def test_filter_recordHasOplogAttribute_true(self):
        mock_record = Mock(spec=logging.LogRecord)
        mock_record.oplog = Mock()
        filter = OperationLogFilter()
        self.assertTrue(filter.filter(record=mock_record))
        


# Get the parent directory of the current file (project_demo folder)
import logging
import os
from pathlib import Path
import sys
import time

current_dir = os.path.dirname(os.path.abspath(__file__))
# Get the path to the repository root directory (one level up)
repository_root = os.path.abspath(os.path.join(current_dir, "..", ".."))

# Add the repository root directory to the Python path
sys.path.append(repository_root)

from oplog import Operated, Operation, OperationHandler # noqa: E402
from oplog.formatters import OplogCsvFormatter # noqa: E402


csv_op_handler = OperationHandler(
    handler=logging.FileHandler(filename=Path("oplogs.csv")), # <-- any logging handler
    formatter=OplogCsvFormatter(), # <-- use your own custom formatter, or built-in ones
)
logging.basicConfig(level=logging.INFO, handlers=[csv_op_handler])

class FluentCalculator:
    @Operated()
    def __init__(self):
        self.value = 0

    @Operated()
    def add(self, num):
        time.sleep(1)
        self.value += num
        return self

    @Operated()
    def subtract(self, num):
        time.sleep(1)
        self.value -= num
        return self

    @Operated()
    def divide(self, num):
        time.sleep(1)
        self.value /= num
        return self
    
    @Operated()
    def get_result(self):
        time.sleep(1)
        return self.value

with Operation("first_calc"):
    result1 = FluentCalculator().add(5).subtract(3).get_result()
    
with Operation("second_calc", suppress=True):
    result2 = FluentCalculator().add(5).divide(0).get_result()




import logging
from typing import Optional

from oplog import OperationLogFilter


class OperationHandler(logging.Handler):
    def __init__(self, 
                 handler: logging.Handler, 
                 formatter: Optional[logging.Formatter] = None, 
                 *args, 
                 **kwargs):
        """A logging handler that only handles operation logs.

        Args:
            handler (logging.Handler): _description_
            formatter (Optional[logging.Formatter], optional): _description_.
                If None, the handler's formatter will be used.
        """
        super().__init__(*args, **kwargs)
        self.handler = handler
        self.addFilter(OperationLogFilter())
        if formatter:
            self.handler.setFormatter(formatter)

    def emit(self, record):
        self.handler.emit(record)


from functools import update_wrapper
from typing import Optional
from oplog.operation import Operation


class Operated:
    def __init__(self, name: Optional[str] = None, suppress: bool = False):
        """Creates an underlying operation. Refer to Operation for more details.
        This should be used as a decorator on a function or method.
        For arguments documentation, refer to Operation.__init__.
        """
        self.name = name
        self.suppress = suppress

    def __call__(self, func):
        def wrapper(*args, **kwargs):
            if self.name is not None:
                op_name = self.name
            else:
                function_name = func.__name__
                try:
                    # get class name if possible
                    qualifier_name = func.__qualname__.split(".")[0]
                except AttributeError:
                    # otherwise, get module name
                    qualifier_name = func.__module__

                op_name = f"{qualifier_name}.{function_name}"
                
            with Operation(name=op_name, suppress=self.suppress):
                result = func(*args, **kwargs)

            return result

        update_wrapper(wrapper, func)
        return wrapper



import inspect
import logging
from unittest.mock import patch
from parameterized import parameterized
from oplog.exceptions import (
    GlobalOperationPropertyAlreadyExistsException,
    OperationPropertyAlreadyExistsException
)
from oplog.operation import Operation

from oplog.tests.logged_test_case import OpLogTestCase


class OperationExceptionTest(Exception):
    pass


class TestOperation(OpLogTestCase):
    def test_operation_loggerLogCalled(self):
        with patch.object(
            logging.Logger,
            logging.Logger.log.__name__,
            return_value=None) as mock_log:
            with Operation(name="test_op") as op:
                pass

        mock_log.assert_called_once_with(
            level=logging.getLevelName(op.log_level),
            msg="operation logged",
            extra={"oplog": op},
        )
        
    def test_operation_loggerIsOfTheCaller(self):
        expected_logger = logging.getLogger(inspect.getmodule(TestOperation).__name__)
        
        with Operation(name="test_op") as op:
                pass
        
        actual_logger = op._logger
        self.assertEqual(expected_logger, actual_logger)

    def test_operation_contextManagerLogged(self):
        with Operation(name="test_op"):
            pass

        assert len(self.ops) == 1

    @parameterized.expand(
        [
            ("Null Type", None),
            ("Bool Type", True),
            ("Int Type", 1),
            ("Float Type", 1.5),
            ("String type", "1"),
        ]
    )
    def test_add_customPropAdded(self, name, value):
        prop_name = "test_custom_prop"
        prop_value = value

        with Operation(name="test_op") as op:
            op.add(prop_name=prop_name, value=prop_value)

        self.assertEqual(len(self.ops), 1)

        op = self.get_op("test_op")
        self.assertIn(prop_name, op.custom_props)
        self.assertEqual(op.custom_props[prop_name], prop_value)

    def test_addGlobal_customPropAdded(self):
        prop_name = "test_global_prop"
        prop_value = 1
        Operation.add_global(prop_name=prop_name, value=prop_value)

        with Operation(name="test_op") as op:
            op.add(prop_name=prop_name, value=prop_value)

        self.assertEqual(len(self.ops), 1)

        op = self.get_op("test_op")
        self.assertIn(prop_name, op.global_props)
        self.assertEqual(op.global_props[prop_name], prop_value)

    def test_addGlobal_globalPropertyAlreadyExists_raises(self):
        prop_name = "test_global_prop"
        prop_value_1 = 1
        prop_value_2 = 2
        
        with self.assertRaises(GlobalOperationPropertyAlreadyExistsException):
            Operation.add_global(prop_name=prop_name, value=prop_value_1)
            Operation.add_global(prop_name=prop_name, value=prop_value_2)
            
    def test_addGlobal_unsupportedType_raises(self):
        prop_name = "test_global_prop"
        prop_value = { "test": 1 }
        
        with self.assertRaises(TypeError):
            Operation.add_global(prop_name=prop_name, value=prop_value)

    def test_operation_add_valueIsDict_customPropsAdded(self):
        prop_name_1 = "test_custom_prop_1"
        prop_name_2 = "test_custom_prop_2"
        prop_value_1 = 1
        prop_value_2 = 2
        prop_bag = {
            prop_name_1: prop_value_1,
            prop_name_2: prop_value_2,
        }

        with Operation(name="test_op") as op:
            op.add_multi(bag=prop_bag)

        self.assertEqual(len(self.ops), 1)

        op = self.get_op("test_op")
        self.assertIn(prop_name_1, op.custom_props)
        self.assertEqual(op.custom_props[prop_name_1], prop_value_1)
        self.assertIn(prop_name_2, op.custom_props)
        self.assertEqual(op.custom_props[prop_name_2], prop_value_2)

    def test_operation_addMulti_customPropsAdded(self):
        prop_name_1 = "test_custom_prop_1"
        prop_name_2 = "test_custom_prop_2"
        prop_value_1 = 1
        prop_value_2 = 2
        prop_bag = {
            prop_name_1: prop_value_1,
            prop_name_2: prop_value_2,
        }
        prop_bag_name = "test_bag"

        with Operation(name="test_op") as op:
            op.add(prop_name=prop_bag_name, value=prop_bag)

        self.assertEqual(len(self.ops), 1)

        op = self.get_op("test_op")
        self.assertIn(prop_bag_name, op.custom_props)
        self.assertEqual(op.custom_props[prop_bag_name], prop_bag)

    def test_operation_exceptionThrownWitoutHandling_failedOperationLoggedExceptionReraised(self):  # noqa: E501
        with self.assertRaises(OperationExceptionTest):
            with Operation(name="test_op") as op:
                raise OperationExceptionTest("test exception")

        self.assertEqual(len(self.ops), 1)

        op = self.get_op("test_op")
        self.assertEqual(op.exception_type, "OperationExceptionTest")
        self.assertEqual(op.exception_msg, "test exception")
    
    def test_operation_exceptionThrownWitoutSuppression_failedOperationLoggedExceptionReraised(self):  # noqa: E501
        with self.assertRaises(OperationExceptionTest):
            with Operation(name="test_op", suppress=False) as op:
                raise OperationExceptionTest("test exception")

        self.assertEqual(len(self.ops), 1)

        op = self.get_op("test_op")
        self.assertEqual(op.exception_type, "OperationExceptionTest")
        self.assertEqual(op.exception_msg, "test exception")

    def test_operation_exceptionThrownWithSuppression_failedOperationLoggedFlowContinues(self):  # noqa: E501
        try:
            with Operation(name="test_op", suppress=True) as op:
                # obvious if to help IDE with "unreachable code"
                if (1 == 1):
                    raise OperationExceptionTest("test exception")
        except OperationExceptionTest:
            # for making this code intentional and explicit
            self.fail("exception should have been suppressed")

        self.assertEqual(len(self.ops), 1)

        op = self.get_op("test_op")
        self.assertEqual(op.exception_type, "OperationExceptionTest")
        self.assertEqual(op.exception_msg, "test exception")
        
    def test_operation_failureWithSuppression_exceptionLogged(self):
        with Operation(name="test_op", suppress=True) as op:
            raise OperationExceptionTest("test exception")

        self.assertEqual(len(self.ops), 1)

        op = self.get_op("test_op")
        self.assertEqual(op.exception_type, "OperationExceptionTest")
        self.assertEqual(op.exception_msg, "test exception")

    def test_operation_add_propertyExists_exceptionRaised(self):
        prop_name = "test_custom_prop"

        with self.assertRaises(OperationPropertyAlreadyExistsException):
            with Operation(name="test_op") as op:
                op.add(prop_name=prop_name, value=1)
                op.add(prop_name=prop_name, value=2)

    def test_operation_parentOperation_parentAndChildrenSet(self):
        with Operation(name="parent_op"):
            with Operation(name="child_op1"):
                pass

            with Operation(name="child_op2"):
                pass

        self.assertEqual(len(self.ops), 3)

        child_op1 = self.get_op("child_op1")
        child_op2 = self.get_op("child_op2")
        parent_op = self.get_op("parent_op")

        self.assertEqual(child_op1.parent_op, parent_op)
        self.assertEqual(child_op2.parent_op, parent_op)

        self.assertIn(child_op1, parent_op.child_ops)
        self.assertIn(child_op2, parent_op.child_ops)

    def test_operation_correlationId_correlationIdSet(self):
        with Operation(name="parent_op"):
            with Operation(name="child_op1"):
                pass

            with Operation(name="child_op2"):
                pass

        self.assertEqual(len(self.ops), 3)

        child_op1 = self.get_op("child_op1")
        child_op2 = self.get_op("child_op2")
        parent_op = self.get_op("parent_op")

        self.assertEqual(child_op1.correlation_id, parent_op.correlation_id)
        self.assertEqual(child_op2.correlation_id, parent_op.correlation_id)



import logging
from typing import List
import unittest

from oplog.operation import Operation


class ListLoggingHandler(logging.Handler):
    def __init__(self):
        super().__init__()
        self.logs = []

    def emit(self, record):
        self.logs.append(record)


class LoggedTestCase(unittest.TestCase):
    def setUp(self):
        super().setUp()

        logger = logging.getLogger()
        logger.setLevel(logging.DEBUG)
        self.handler = ListLoggingHandler()
        self.handler.logs = []

        logger.addHandler(self.handler)

    def tearDown(self):
        super().tearDown()


class OpLogTestCase(LoggedTestCase):
    @property
    def ops(self) -> List[Operation]:
        return [log.oplog for log in self.handler.logs]

    def get_op(self, name: str) -> Operation:
        for op in self.ops:
            if op.name == name:
                return op
        raise LookupError(f"Operation with name {name} not found")

    def tearDown(self):
        Operation.factory_reset()
        return super().tearDown()

    
class LoggedAsyncTestCase(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        await super().asyncSetUp()

        logger = logging.getLogger()
        logger.setLevel(logging.DEBUG)
        self.handler = ListLoggingHandler()
        self.handler.logs = []

        logger.addHandler(self.handler)
    
    async def asyncTearDown(self):
        await super().asyncTearDown()


class OpLogAsyncTestCase(LoggedAsyncTestCase):
    @property
    def ops(self) -> List[Operation]:
        return [log.oplog for log in self.handler.logs]

    def get_op(self, name: str) -> Operation:
        for op in self.ops:
            if op.name == name:
                return op
        raise LookupError(f"Operation with name {name} not found")

    async def asyncTearDown(self):
        Operation.factory_reset()
        return super().tearDown()


class OperationPropertyAlreadyExistsException(Exception):
    def __init__(self, op_name: str, prop_name: str) -> None:
        super().__init__(f'property `{prop_name}` already exists in op `{op_name}`')
        
class GlobalOperationPropertyAlreadyExistsException(Exception):
    def __init__(self, prop_name: str) -> None:
        super().__init__(f'global property `{prop_name}` already exists')
        
class LogRecordMissingOperationException(Exception):
    pass


from .operation import Operation  # noqa: F401
from .operated import Operated  # noqa: F401
from .operation_log_filter import OperationLogFilter  # noqa: F401
from .operation_handler import OperationHandler  # noqa: F401